/*
  18bDAC, from LeoLed
  Uses LTC2756 DAC and Isolated SPI
  D. Erickson, Erickson Engineering
  9/30/17  Started change from LCD to OLED using u8g2 library. Very nice.
  10/1/17  Added keys and encoder
  10/7/17  Built this version with 10b ADC, no I2C or isolation.  2/13/20  Started SPI
  3/16/17  Adding interrupts to Timer0 using technique on Adafruit web page
          https://learn.adafruit.com/multi-tasking-the-arduino-part-2/timers
  2/13/20  SPI, Bit-bang 32 bits for DAC started
  2/15/20  Fixing bug with key variable> After encoder, repeats key
  3/1/20   Changed to integer Iset, in uV so < 1LSB increment. This is until 1.048V reference
           Range LSB       Step
            20V, 76uV, use 50uV
            10V, 38uV, use 20uV
            5V,  19uV, use 10uV
           Still need to re-burn boot after every download!!!
   3/28/21 10.48V Reference, 40uV / LSB
   3/30/21 Re-factored DAC and display, Adding SCPI 
   4/1/21  SCPI Set V and Read V working with OIC SCPI library!    
            Needs range set, scaling and CAL per range. Although DAC is very good. 
   4/19/21 SCPI library changed to Verkrer Library. Works fine. 
           Iset now rounds to nearest 40uV value
              
*/


#include <Wire.h>
#include <stdio.h>
#include <U8g2lib.h>

#include "Vrekrer_scpi_parser.h"

#ifdef U8X8_HAVE_HW_SPI
#include <SPI.h>
#endif
#ifdef U8X8_HAVE_HW_I2C
#include <Wire.h>
#endif

// From U8g2 Contructor List (Frame Buffer)
U8G2_SH1106_128X64_NONAME_F_4W_HW_SPI u8g2(U8G2_R0, /* cs=*/ 10, /* dc=*/ 9, /* reset=*/ 8);

#define DIG 5  //Number of digits after decimal point
#define FPWID 6

// 18b DAC scaling
#define DAC_REF 5.243
#define ISET_MAX 10400000L      // Can be as high as 10485000 10.4V is more convenient 
#define DAC_MAX ISET_MAX/40L

// Pin Definitions
#define TESTPIN 0       // Just for test point
// Encoder
#define ENCA 4
#define ENCB 7
//#define S1 A4   // Reading S1 doesn't work with D17 uses nonstandard pin
// Switches
//#define S1 17
#define S2 11     // typically RIGHT
#define S3 12     // Typically LEFT
#define S4 5
#define S5 13

// Pins for SPI to DAC
#define SPI_CK    A5
#define SPI_MOSI  A4
#define SPI_SS    A2

// DAC ranges
#define RU5V   0x000L
#define RU10V  0x100L
#define RB5V   0x200L
#define RB10V  0x300L
#define RB2_5V 0x400L
#define RB2_7V 0x500L

/* Encoder action codes */
#define NADA 0
#define INCR 1
#define DECR 2
#define ILL 3
/* Set to the appropriate encoder used */
#define ENC4PPCLICK
//#define ENC1PPCLICK

// Key state, to debounce, etc. one for each key. Also KeyFlag, set by interrupts
#define NUMKEYS 4
#define KEYIDLE 0
#define KEYVALID 20
#define KEYWAIT KEYVALID + 1

#define NUM_DISP 4
#define MAX_DISP NUM_DISP -1

#define DIGIT_MAX 0
#define DIGIT_MIN -5

/*  Global Variables */
SCPI_Parser my_instrument;
double sVoltage = 0.0L;
double voltage = 123.456789L;
double current = 0.000123456L;

unsigned int count = 0;
char chan;        // ADC Channel
float fIset;
long Iset;        // Current setting in uV
//int range;        // DAC range
long DacBits;

/* Encoder and Keypad stuff */
static unsigned char keyState[NUMKEYS];
static unsigned char keyFlag[NUMKEYS];
char command;
int digit;
int16_t encoder_count = 0;      /* Number of encoder counts since last read */
int16_t old_encoder = 0;        /* Previous value */




/****************************** setup() runs once *******************************/
void setup() {
  int i;
  
  initScpi();
  
  // Initialize serial communications
  Serial.begin(115200);

  fIset = 0.0;
  Iset = 0;
  
  digit = -1;     // Start with tenth's digit 
  u8g2.begin();

  u8g2_prepare();
  // Splash screen
  u8g2.drawStr( 0, 10, "18b DAC r2.0");
  u8g2.drawStr( 0, 20, "Leoled + Vrekrer SCPI");
  u8g2.drawStr( 0, 30, "Erickson Engineering");
  u8g2.sendBuffer();
  delay(3000);          // 3S delay to show startup and to allow USB Serial port to connect
  Serial.println("18bDAC By D. Erickson");

  /* Setup for a user Timer0 1ms interrupt. Default time is 16MHz/256/64 = 976.6 Hz
     Timer0 is already used for millis() - we'll interrupt about the middle and call the
     "Compare A" function below. OCR0A is not uised by Arduino timers 
     NOTE: This is AVR MEGA32xx specific code */
  OCR0A = 0x128;           // Set to 128 out of 255 or 50%. Doesn't make much difference. 
  TIMSK0 |= _BV(OCIE0A);  // Set interrupt enable
// Set up I/O pins for LED, encoder and switches
  pinMode(TESTPIN, OUTPUT);
  pinMode(ENCA, INPUT_PULLUP);     // two encoder pins
  pinMode(ENCB, INPUT_PULLUP);
  pinMode(S2, INPUT_PULLUP);
  pinMode(S3, INPUT_PULLUP);
  pinMode(S4, INPUT_PULLUP);
  pinMode(S5, INPUT_PULLUP);

  pinMode(SPI_CK, OUTPUT);   // Set up SPI pins, no MISO for now
  pinMode(SPI_MOSI, OUTPUT);
  pinMode(SPI_SS, OUTPUT);
  digitalWrite(SPI_CK, 0);
  digitalWrite(SPI_MOSI, 0);
  digitalWrite(SPI_SS, 1);

  for (i = 0; i < NUMKEYS; i ++) keyFlag[i] = 0; // Initialize keypad flags

}

/************************ loop() runs forever ***********************/
void loop()
{
  char key;
  //char incomingByte;

  char line_buffer[128];
  unsigned char read_length;

  /* Original serial stuff. Not used with SCPI
  // Get control, act on it.
  if (Serial.available() > 0)
  {
    // A while loop here would read faster than one byte per loop.
    // read the incoming byte:
    incomingByte = Serial.read(); // read one byte from serial
    command = incomingByte;
    count = 1;
    Serial.print("Ser: ");
    Serial.println(command);
  }
  */
  
  // Get Encoder changes.
  count = 0;
  
  if (encoder_count != old_encoder)
  {
    if (encoder_count > old_encoder)
    {
      command = 'U';
      count = encoder_count - old_encoder;
      old_encoder = encoder_count;
      //Serial.println("Enc: U");
    }
    if (encoder_count < old_encoder)
    {
      command = 'D';
      count =  old_encoder - encoder_count ;
      old_encoder = encoder_count;
      //Serial.println("Enc: D");
    }
  }

  key = getKeypad();
  if (key)
  {
    command = key;
    //Serial.print("Key: ");
    //Serial.println(command);
  }
  
  switch (command)
  {
    case 'U':    
      if (digit ==  0) Iset += (long)count * 1000000l;
      if (digit == -1) Iset += (long)count * 100000l;
      if (digit == -2) Iset += (long)count * 10000l;
      if (digit == -3) Iset += (long)count * 1000l;
      if (digit == -4) Iset += (long)count * 100l;
      if (digit == -5) Iset += (long)count * 40l;       // 40uv / LSB
      setDac();
      // Serial.println("U");
      break;
      
    case 'D':
      if (digit ==  0) Iset -= (long)count * 1000000l;
      if (digit == -1) Iset -= (long)count * 100000l;
      if (digit == -2) Iset -= (long)count * 10000l;
      if (digit == -3) Iset -= (long)count * 1000l;
      if (digit == -4) Iset -= (long)count * 100l;
      if (digit == -5) Iset -= (long)count * 40l;
      setDac();
      // Serial.println("D");
      break;   
    
    case 'l':    // move digit left
      digit = digit + 1;
      if  (digit > DIGIT_MAX) digit = DIGIT_MAX;
      //Serial.println("l");
      //Serial.println(digit);
      break;
      
    case 'r':
      digit = digit - 1;
      if  (digit < DIGIT_MIN) digit = DIGIT_MIN;
      //Serial.println(command);
      //Serial.println(digit);
      break;
      
    default:
      break;
  }
  command = 0;
  setDac();
  updateDisp();
  
  /* SCPI: Read in a line and execute it. */
  my_instrument.ProcessInput(Serial, "\n");
  
  //delay(20);        // delay sets display and loop() update rate
}
/*************************** End of loop() ********************/


void setDac(void)
{
  // Set DAC span
  spi32W(0x20000000L | RU10V);   
  // clip and Round Iset and DacBits   
  Iset = constrain(Iset, 0, ISET_MAX);
  Iset = (Iset + 20L) / 40L;             // Set Iset to closest DAC value, with rounding. 
  Iset *= 40L;
//DacBits = (long)(((float)Iset * (float)DAC_MAX / (float)ISET_MAX));   // 10.000 = 250,000
  DacBits = Iset / 40L;   // Since Iset is clipped and rounded, this works
  if (DacBits < 0) {
    DacBits = 0L;
    Iset = 0L;
  }
  if (DacBits > DAC_MAX) {
    DacBits = DAC_MAX;
  }
  
  // Set the DAC
  //  Serial.println (DacBits, HEX);
  spi32W(0x70000000L | (DacBits << 6));      // Write Code and update: Dac bits 23:6
}

void setRange(int range)  // Currently unused
{
  // Set DAC span (range)
  spi32W(0x20000000L | range);      // Write Span
}


void updateDisp(void)
{
  char buf1[20];
  // Display on OLED
  u8g2_prepare();
// First Line
  u8g2.drawStr( 0, 0, "18b DAC r2.0 ");
//  dtostrf(fIset, 9, 5, buf1);
  ltoa (DacBits, buf1, 10);  // Display DacBits
  u8g2.drawStr( 0, 8, buf1);

// Big Font
  u8g2.setFont(u8g2_font_profont22_mr);
  dtostrf(((float)Iset / 1000000.0), 9, 5, buf1);
  u8g2.drawStr(0, 20, buf1);

// Draw an underline cursor under the selected digit 
  int curPosX;
  int curPosY = 42;   // Under the voltage 
  int curWid = 12;    // Approx character width
  if (digit == 0 ) curPosX = 2 ;
  if (digit <= -1) curPosX = 3 - digit;  //-1 to -5: 4 to 8
  curPosX = curPosX * curWid -1;
  u8g2.drawLine(curPosX, curPosY, curPosX + curWid , curPosY);
  u8g2.sendBuffer();  
}


//Interrupt is called about once per millisecond. Uses Arduino timer 0 
SIGNAL(TIMER0_COMPA_vect)
{
// Test by pulsing TESTPIN High
//  digitalWrite(TESTPIN, 1);
  getEncoder();
  getKeys();
//  digitalWrite(TESTPIN, 0);
}

void u8g2_prepare(void) {
  u8g2.clearBuffer();
  u8g2.setFont(u8g2_font_6x10_tr);
  u8g2.setFontRefHeightExtendedText();
  u8g2.setDrawColor(1);
  u8g2.setFontPosTop();
  u8g2.setFontDirection(0);
  u8g2.setContrast(255);
}

/* Encoder */

/* For encoders that generate 4 transitions per click such as most mechanical encoders.
   Note that mechanical encoders need pull up resistors and should also have R-C filters
*/
const  uint8_t encoder_logic[] = {
  /*old new      action
    ---------------------- */
  /* 00 00 */    NADA,
  /* 00 01 */    NADA,
  /* 00 10 */    DECR,
  /* 00 11 */    NADA,
  /* 01 00 */    NADA,
  /* 01 01 */    NADA,
  /* 01 10 */    NADA,
  /* 01 11 */    NADA,

  /* 10 00 */    INCR,
  /* 10 01 */    NADA,
  /* 10 10 */    NADA,
  /* 10 11 */    NADA,
  /* 11 00 */    NADA,
  /* 11 01 */    NADA,
  /* 11 10 */    NADA,
  /* 11 11 */    NADA
};

/* getEncoder()

   This is the interrupt routine for the encoder.
   The encoder is an optical encoder knob that outputs A-B quadrature signals. The knob is
   rotated relatively slowly, so once every 1 mS the 2 bits are polled. These are compared
   to the previous readings and depending on the sequence, either the count value is
   incremented, decremented or not action (NADA).
   The logic is contained in encoder_logic[] a table.
   Here's an example:
    old new action
    32  10
   --------------------
    00  00 no action
    00  01 increment
    ...
   The encoder state is 4 bits. The 2 LSBs are the new bits. The MSBs [3:2] are the previous readings.
*/
void getEncoder(void)
{
  uint8_t action;
  static uint8_t encoder_state;
  /* read 2 LSBs, shift up the previous bits */
  encoder_state = (encoder_state << 2) | (digitalRead(ENCB) << 1) | digitalRead(ENCA);
  action = encoder_logic[encoder_state & 0x0F];   /* Look up action: 4 LSBs only */
  if (action == INCR)
    encoder_count++;
  if (action == DECR)
    encoder_count--;
}


/* Interrupt to debounce keys. If the key is down, increment keyState. If not, reset to KEYIDLE, 0
    This contains the mapping between pins and keyState[]
    Set flag if count resches KEYVALID, thn hold until key is released.
*/

void getKeys (void)
{
  int keyIndex;
  // Read all the keys. Increment or reset keyState[]
  
  //  if (digitalRead(S1) == 0) keyState[0]++;
  //  else keyState[0] = KEYIDLE;

  if (digitalRead(S2) == 0) keyState[1]++;
  else keyState[1] = KEYIDLE;

  if (digitalRead(S3) == 0) keyState[2]++;
  else keyState[2] = KEYIDLE;

  if (digitalRead(S5) == 0) keyState[3]++;
  else keyState[3] = KEYIDLE;
  
  for (keyIndex = 0; keyIndex < NUMKEYS; keyIndex++)
  {
    if (keyState[keyIndex] >= KEYWAIT)
    {
      keyState[keyIndex] = KEYWAIT;  // Hold in KEYWAIT
    }
    if (keyState[keyIndex] == KEYVALID)             // Generate a key flag.
    {
      keyFlag[keyIndex] = 1;
      keyState[keyIndex] = KEYWAIT;
    }
  }
}

/* Read the key flags set by interrupts
   Map to the key codes
   keyFlag[0] = 't'
   keyFlag[2]S3
   Put the key in inKey[] and set the flag */
char getKeypad(void)
{
  char key2 = 0;
  
  if (keyFlag[0])
  {
    key2 = 't';
    keyFlag[0] = 0;
    //Serial.println("S1 Pressed");
  }
  if (keyFlag[1])
  {
    key2 = 'r';
    keyFlag[1] = 0;
    //Serial.println("S2 Pressed");
  }
  if (keyFlag[2])
  {
    key2 = 'l';
    keyFlag[2] = 0;
    //Serial.println("S3 Pressed");
  }
  if (keyFlag[3])
  {
    key2 = 's';
    keyFlag[3] = 0;
    //Serial.println("S5 Pressed");
  }
  //  Serial.println(keyState[0], DEC);
  //  Serial.println(keyState[1], DEC);
  //  Serial.println(keyState[2], DEC);
  //  Serial.print("In getKeypad() key2 = ");
  //  Serial.println(key2, DEC);
  return key2;

}


/*  Low-level Bit-bang 32 bit SPI Write for LTC2756 DAC.
    Data is clocked in on + edge of CK\
    Assert SS/, clock out 32 bits, de-assert /SS
*/
void spi32W(long data)
{
  int i;
  digitalWrite(SPI_CK, 0);    // Should already be in this state
  digitalWrite(SPI_MOSI, 0);
  digitalWrite(SPI_SS, 1);

  digitalWrite(SPI_SS, 0);
  for (i = 0; i < 32; i++)
  {
    if (data & 0x80000000l ) digitalWrite(SPI_MOSI, 1);  // MOSI = 1
    else digitalWrite(SPI_MOSI, 0);                 // MOSI = 0
    digitalWrite(SPI_CK, 1);
    digitalWrite(SPI_CK, 0);
    data = data << 1;                               // Shift up
  }
  digitalWrite(SPI_SS, 1);
}
